@@ -24,7 +24,6 @@ import os
 import sys
 import re
 import traceback
-import math
 
 import pyotherside
 
@@ -32,20 +31,25 @@ try:
     from StringIO import StringIO # python 2
 except ImportError:
     from io import StringIO # python 3
-from pprint import pprint
 
 # modRana imports
 import math
 from modules.gui_modules.base_gui_module import GUIModule
-from datetime import datetime
 import time
+import threading
 
-from core.fix import Fix
 from core import signal
-from core.backports import six
 from core import constants
 from core.threads import threadMgr
+from core import geo
+from core import modrana_log
 
+import logging
+log = logging.getLogger("mod.gui.qt5")
+qml_log = logging.getLogger("mod.gui.qt5.qml")
+
+SEARCH_STATUS_PREFIX = "search:status:"
+SEARCH_RESULT_PREFIX = "search:result:"
 
 def newlines2brs(text):
     """ QML uses <br> instead of \n for linebreak """
@@ -54,6 +58,21 @@ def newlines2brs(text):
 def getModule(m, d, i):
     return QMLGUI(m, d, i)
 
+def point2dict(point):
+    """ Convert a Point instance to a dict
+
+        :param Point point: a Point object instance
+        :returns dict: a dict representation of the point
+    """
+    return {
+        "name" : point.name,
+        "description" : point.description,
+        "latitude" : point.lat,
+        "longitude" : point.lon,
+        "elevation" : point.elevation,
+        "highlight" : False
+    }
+
 
 class QMLGUI(GUIModule):
     """A Qt + QML GUI module"""
@@ -66,8 +85,12 @@ class QMLGUI(GUIModule):
         self.centeringDisableThreshold = 2048
         self.firstTimeSignal = signal.Signal()
         size = (800, 480) # initial window size
+
         # register exit handler
-        pyotherside.atexit(self._shutdown)
+        #pyotherside.atexit(self._shutdown)
+        # FIXME: for some reason the exit handler is never
+        # called on Sailfish OS, so we use a onDestruction
+        # handler on the QML side to trigger shutdown
 
         # window state
         self._fullscreen = False
@@ -75,11 +98,11 @@ class QMLGUI(GUIModule):
         # get screen resolution
         # TODO: implement this
         #screenWH = self.getScreenWH()
-        #print(" @ screen size: %dx%d" % screenWH)
+        #self.log.debug(" @ screen size: %dx%d" % screenWH)
         #if self.highDPI:
-        #    print(" @ high DPI")
+        #    self.log.debug(" @ high DPI")
         #else:
-        #    print(" @ normal DPI")
+        #    self.log.debug(" @ normal DPI")
 
         # NOTE: what about multi-display devices ? :)
 
@@ -89,6 +112,10 @@ class QMLGUI(GUIModule):
             "icon" : IconImageProvider(self),
             "tile" : TileImageProvider(self),
         }
+
+        with open("data/tile_not_found.png", "rb") as f:
+            self._tileNotFoundImage = f.read()
+
         ## register the actual callback, that
         ## will call the appropriate provider base on
         ## image id prefix
@@ -110,21 +137,25 @@ class QMLGUI(GUIModule):
         # provides easy access to modRana modules from QML
         self.modules = Modules(self)
 
-    def firstTime(self):
-        self._location = self.m.get('location', None)
-        self._mapTiles = self.m.get('mapTiles', None)
-        self._mapLayers = self.m.get('mapLayers', None)
+        # search functionality for the QML context
+        self.search = Search(self)
 
+        # make the log manager easily accessible
+        self.log_manager = modrana_log.log_manager
+
+        # log for log messages from the QML context
+        self.qml_log = qml_log
+
+    def firstTime(self):
         # trigger the first time signal
         self.firstTimeSignal()
 
     def _shutdown(self):
         """Called by PyOtherSide once the QML side is shutdown.
         """
-        print("Qt5 GUI: shutting down")
+        self.log.info("Qt 5 GUI module shutting down")
         self.modrana.shutdown()
 
-
     def getIDString(self):
         return "Qt5"
 
@@ -162,7 +193,7 @@ class QMLGUI(GUIModule):
 
         #    # QML uses <br> instead of \n for linebreak
         #    text = newlines2brs(text)
-        #    print("QML GUI notify:\n message: %s, timeout: %d" % (text, msTimeout))
+        #    self.log.debug("notify:\n message: %s, timeout: %d" % (text, msTimeout))
         #    if self.rootObject:
         #      self.rootObject.notify(text, msTimeout)
         #    else:
@@ -196,11 +227,30 @@ class QMLGUI(GUIModule):
 
     def setPosition(self, posDict):
         lat, lon = float(posDict["latitude"]), float(posDict["longitude"])
-        self.set("pos", (lat, lon))
         elevation = float(posDict["elevation"])
+        metersPerSecSpeed = float(posDict["speedMPS"])  # m/s
+        # report that we have 3D fix
+        # (looks like we can't currently reliably discern between 2D
+        # and 3D fix on the Jolla, might be good to check what other
+        # Sailfish OS running devices report)
+        self.set("fix", 3)
+
+        self.set("pos", (lat, lon))
+
         # check if elevation is valid
         if not math.isnan(elevation):
             self.set("elevation", elevation)
+        else:
+            self.set("elevation", None)
+
+        # check if speed is valid
+        if not math.isnan(metersPerSecSpeed):
+            self.set("speed", metersPerSecSpeed*3.6)
+            self.set("metersPerSecSpeed", metersPerSecSpeed)
+        else:
+            self.set("speed", None)
+            self.set("metersPerSecSpeed", None)
+
         # update done
         self.set('locationUpdated', time.time())
         # TODO: move part of this to the location module ?
@@ -208,52 +258,186 @@ class QMLGUI(GUIModule):
     def _selectImageProviderCB(self, imageId, requestedSize):
         originalImageId = imageId
         providerId = ""
-        #print("SELECT IMAGE PROVIDER")
-        #print(imageId)
-        #print(imageId.split("/", 1))
+        #self.log.debug("SELECT IMAGE PROVIDER")
+        #self.log.debug(imageId)
+        #self.log.debug(imageId.split("/", 1))
         try:
             # split out the provider id
             providerId, imageId = imageId.split("/", 1)
             # get the provider and call its getImage()
             return self._imageProviders[providerId].getImage(imageId, requestedSize)
         except ValueError:  # provider id missing or image ID overall wrong
-            print("Qt5 GUI: provider ID missing: %s" % originalImageId)
+            self.log.error("provider ID missing: %s", originalImageId)
         except AttributeError:  # missing provider (we are calling methods of None ;) )
             if providerId:
-                print("Qt5 GUI: image provider for this ID is missing: %s" % providerId)
+                self.log.error("image provider for this ID is missing: %s", providerId)
             else:
-                import sys
-                e = sys.exc_info()[1]
-                print("Qt5 GUI: image provider broken, image id: %s" % originalImageId)
-                print(e)
+                self.log.error("image provider broken, image id: %s", originalImageId)
         except Exception:  # catch and report the rest
-            import sys
-            e = sys.exc_info()[1]
-            print("Qt5 GUI: image loading failed, imageId: %s" % originalImageId)
-            print(e)
+            self.log.exception("image loading failed, imageId: %s", originalImageId)
+
+    def _tileId2lzxy(self, tileId):
+        """Convert tile id string to the "standard" lzxy tuple
+
+        :param str tileId: map instance name/layer id/z/x/y
+
+        :returns: lzxy tuple
+        :rtype: tuple
+        """
+        split = tileId.split("/")
+        # pinchMapId = split[0]
+        layerId = split[1]
+        z = int(split[2])
+        x = int(split[3])
+        y = int(split[4])
+        # TODO: local id:layer cache ?
+        layer = self.modules.mapLayers.getLayerById(layerId)
+        return layer, z, x, y
+
+    def addTileDownloadRequest(self, tileId):
+        """Add an asynchronous download request, the tile will be
+        notified once the download is finished or fails
+        :param string tileId: unique tile id
+        """
+        try:
+            lzxy = self._tileId2lzxy(tileId)
+            self.modules.mapTiles.addTileDownloadRequest(lzxy, tileId)
+        except Exception:
+            self.log.exception("adding tile download request failed")
+
+    def _getStartupValues(self):
+        """ Return a dict of values needed by the Qt 5 GUI right after startup.
+        By grouping the requested values in a single dict we reduce the number
+        of Python <-> QML roundtrips and also make it possible to more easily
+        get these values asynchronously (values arrive all at the same time,
+        not in random order at random time).
+
+        :returns: a dict gathering the requested values
+        :rtype dict:
+        """
+        values = {
+            "modRanaVersion" : self.getModRanaVersion(),
+            "constants" : self.getConstants(),
+            "showQuitButton": self.showQuitButton(),
+            "fullscreenOnly": self.modrana.dmod.fullscreenOnly(),
+            "shouldStartInFullscreen": self.shouldStartInFullscreen(),
+            "needsBackButton": self.modrana.dmod.needsBackButton(),
+            "needsPageBackground": self.modrana.dmod.needsPageBackground(),
+            "lastKnownPos" : self.get("pos", None),
+            "gpsEnabled" : self.get("GPSEnabled", True),
+            "posFromFile" : self.get("posFromFile", None),
+            "nmeaFilePath" : self.get("NMEAFilePath", None),
+            "layerTree" : self.modules.mapLayers.getLayerTree(),
+            "dictOfLayerDicts" : self.modules.mapLayers.getDictOfLayerDicts(),
+            "themesFolderPath" : os.path.abspath(self.modrana.paths.getThemesFolderPath())
+        }
+        return values
 
 
 class Modules(object):
-    """A class that provides access to modRana modules from the QML context"""
+    """A class that provides access to modRana modules from the QML context,
+       using the __getattr__ method so that QML can access all modules dynamically
+       with normal dot notation
+    """
 
     def __init__(self, gui):
         self._info = None
         self._stats = None
+        self._mapLayers = None
+        self._storeTiles = None
         self.gui = gui
 
-    @property
-    def info(self):
-        """A lazy evaluated property providing access to the info module"""
-        if self._info is None:
-            self._info = self.gui.m.get("info")
-        return self._info
-    
-    @property
-    def stats(self):
-        """A lazy evaluated property providing access to the stats module"""
-        if self._stats is None:
-            self._stats = self.gui.m.get("stats")
-        return self._stats
+    def __getattr__(self, moduleName):
+        return self.gui.m.get(moduleName, None)
+
+
+class Search(object):
+    """An easy to use search interface for the QML context"""
+
+    def __init__(self, gui):
+        self.gui = gui
+        self._threadsInProgress = {}
+        # register the thread status changed callback
+        threadMgr.threadStatusChanged.connect(self._threadStatusCB)
+
+    def search(self, searchId, query):
+        """Trigger an asynchronous search (specified by search id)
+        for the given term
+
+        :param str query: search query
+        """
+        online = self.gui.m.get("onlineServices", None)
+        if online:
+            # construct result handling callback
+            callback = lambda x : self._searchCB(searchId, x)
+            # get search function corresponding to the search id
+            searchFunction = self._getSearchFunction(searchId)
+            # start the search and remember the search thread id
+            # so we can use it to track search progress
+            # (there might be more searches in progress so we
+            #  need to know the unique search thread id)
+            threadId = searchFunction(query, callback)
+            self._threadsInProgress[threadId] = searchId
+            return threadId
+
+    def _searchCB(self, searchId, results):
+        """Handle address search results
+
+        :param list results: address search results
+        """
+        # try to sort local search results by distance
+        # TODO: make this configurable
+        if searchId == "local":
+            pos = self.gui.get("pos", None)
+            if pos:
+                distanceList = []
+                for result in results:
+                    distanceList.append((geo.distanceP2LL(result,pos[0], pos[1]), result))
+                distanceList.sort()
+                results = map(lambda x: x[1], distanceList)
+        # covert the Points in the results to a list of dicts
+        resultList = []
+        for result in results:
+            resultList.append(point2dict(result))
+
+        resultId = SEARCH_RESULT_PREFIX + searchId
+        pyotherside.send(resultId, resultList)
+        thisThread = threading.currentThread()
+        # remove the finished thread from tracking
+        if thisThread.name in self._threadsInProgress:
+            del self._threadsInProgress[thisThread.name]
+
+    def cancelSearch(self, threadId):
+        """Cancel the given asynchronous search thread"""
+        log.info("canceling search thread: %s", threadId)
+        threadMgr.cancel_thread(threadId)
+        if threadId in self._threadsInProgress:
+            del self._threadsInProgress[threadId]
+
+    def _threadStatusCB(self, threadName, threadStatus):
+        # check if the event corresponds to some of the
+        # in-progress search threads
+        recipient = self._threadsInProgress.get(threadName)
+        if recipient:
+            statusId = SEARCH_STATUS_PREFIX + recipient
+            pyotherside.send(statusId, threadStatus)
+
+    def _getSearchFunction(self, searchId):
+        """Return the search function object for the given searchId"""
+        online = self.gui.m.get("onlineServices", None)
+        if online:
+            if searchId == "address":
+                return online.geocodeAsync
+            elif searchId == "wikipedia":
+                return online.wikipediaSearchAsync
+            elif searchId == "local":
+                return online.localSearchAsync
+            else:
+                log.error("search function for id: %s not found", searchId)
+                return None
+        else:
+            log.error("onlineServices module not found")
+
 
 class ImageProvider(object):
     """PyOtherSide image provider base class"""
@@ -272,8 +456,8 @@ class IconImageProvider(ImageProvider):
         ImageProvider.__init__(self, gui)
 
     def getImage(self, imageId, requestedSize):
-        #print("ICON!")
-        #print(imageId)
+        #log.debug("ICON!")
+        #log.debug(imageId)
         try:
             #TODO: theme name caching ?
             themeFolder = self.gui.modrana.paths.getThemesFolderPath()
@@ -288,72 +472,98 @@ class IconImageProvider(ImageProvider):
             if not os.path.exists(fullIconPath):
                 if splitPath[0] == constants.DEFAULT_THEME_ID:
                     # already on default theme and icon path does not exist
-                    print("Icon not found in default theme:")
-                    print(fullIconPath)
+                    log.error("Icon not found in default theme:")
+                    log.error(fullIconPath)
                     return None
                 else:  # try to get the icon from default theme
                     splitPath[0] = constants.DEFAULT_THEME_ID
                     fullIconPath = os.path.join(themeFolder, *splitPath)
                     if not os.path.exists(fullIconPath):
                         # icon not found even in the default theme
-                        print("Icon not found even in default theme:")
-                        print(fullIconPath)
+                        log.error("Icon not found even in default theme:")
+                        log.error(fullIconPath)
                         return None
             with open(fullIconPath, 'rb') as f:
                 # the context manager will make sure the icon
                 # file is properly closed
                 return bytearray(f.read()), (-1,-1), pyotherside.format_data
         except Exception:
-            import sys
-            e = sys.exc_info()[1]
-            print("Qt5 GUI: icon image provider: loading icon failed")
-            print(e)
-            print(os.path.join('themes', imageId))
-            print("Traceback:")
-            traceback.print_exc(file=sys.stdout) # find what went wrong
-
+            log.exception("icon image provider: loading icon failed, id:\n%s" % imageId)
 
 class TileImageProvider(ImageProvider):
     """
     the TileImageProvider class provides images images to the QML map element
-    NOTE: this image provider is currently only used as fallback in case
-    the localhost tileserver won't start
     """
 
     def __init__(self, gui):
         ImageProvider.__init__(self, gui)
+        self.gui = gui
+
+        self.gui.firstTimeSignal.connect(self._firstTimeCB)
+
+    def _firstTimeCB(self):
+        # connect to the tile downloaded callback so that we can notify
+        # the QML context that a tile has ben downloaded and should be
+        # shown on the screen
+        # NOTE: we need to wait for the firstTime signal as at GUI module init
+        # the other modules (other than the device module) are not yet initialized
+        self.gui.modules.mapTiles.tileDownloaded.connect(self._tileDownloadedCB)
+
+    def _tileDownloadedCB(self, error, lzxy, tag):
+        """Notify the QML context that a tile has been downloaded"""
+        pinchMapId = tag.split("/")[0]
+        #log.debug("SENDING: %s %s" % ("tileDownloaded:%s" % pinchMapId, tag))
+        pyotherside.send("tileDownloaded:%s" % pinchMapId, tag, error)
 
     def getImage(self, imageId, requestedSize):
         """
         the tile info should look like this:
         layerID/zl/x/y
         """
-        #print("TILE REQUESTED")
-        #print(imageId)
-        #print(requestedSize)
+        #log.debug("TILE REQUESTED %s" % imageId)
+        #log.debug(requestedSize)
         try:
             # split the string provided by QML
             split = imageId.split("/")
-            layerId = split[0]
-            z = int(split[1])
-            x = int(split[2])
-            y = int(split[3])
+            pinchMapId = split[0]
+            layerId = split[1]
+            z = int(split[2])
+            x = int(split[3])
+            y = int(split[4])
+
+            # TODO: local id:layer cache ?
+            layer = self.gui.modules.mapLayers.getLayerById(layerId)
+
+            # construct the tag
+            #tag = (pinchMapId, layerId, z, x, y)
+            #tag = (pinchMapId, layerId, z, x, y)
 
             # get the tile from the tile module
-            tileData = self.gui._mapTiles.getTile(layerId, z, x, y)
-            if not tileData:
-                #print("NO TILEDATA")
-                return None
-            return bytearray(tileData), (256,256), pyotherside.format_data
+            tileData = self.gui.modules.mapTiles.getTile((layer, z, x, y),
+                                                         async=True, tag=imageId,
+                                                         download=False)
+            imageSize = (256,256)
+            if tileData is None:
+                # The tile was not found locally
+                # * in persistent storage (files/sqlite db)
+                # * in the tile cache in memory
+                # An asynchronous tile download request has been added
+                # automatically, so we just now need to notify the
+                # QtQuick GUI that it should wait fo the download
+                # completed signal.
+                #
+                # We notify the GUI by returning a 1x1 image.
+                tileData = self.gui._tileNotFoundImage
+                imageSize = (1,1)
+                #log.debug("%s NOT FOUND" % imageId)
+                # TODO: use some raw image data instead of PNG ?
+            #log.debug("RETURNING STUFF %d %s" % (imageSize[0], imageId))
+            return bytearray(tileData), imageSize, pyotherside.format_data
         except Exception:
-            import sys
-            e = sys.exc_info()[1]
-            print("Qt 5 GUI: tile image provider: loading tile failed")
-            print(e)
-            print(imageId)
-            print(requestedSize)
-            traceback.print_exc(file=sys.stdout)
-
+            log.error("tile image provider: loading tile failed")
+            log.error(imageId)
+            log.error(requestedSize)
+            log.exception()
 
 class MapTiles(object):
     def __init__(self, gui):
@@ -373,52 +583,20 @@ class MapTiles(object):
         True - tile already in storage or in memory
         False - tile download in progress, retry in a while
         """
-        #    print(layerId, z, x, y)
-        if self.gui._mapTiles.tileInMemory(layerId, z, x, y):
-        #      print("available in memory")
+        #    log.debug(layerId, z, x, y)
+        if self.gui.mapTiles.tileInMemory(layerId, z, x, y):
+        #      log.debug("available in memory")
             return True
-        elif self.gui._mapTiles.tileInStorage(layerId, z, x, y):
-        #      print("available in storage")
+        elif self.gui.mapTiles.tileInStorage(layerId, z, x, y):
+        #      log.debug("available in storage")
             return True
         else: # not in memory or storage
             # add a tile download request
-            self.gui._mapTiles.addTileDownloadRequest(layerId, z, x, y)
-            #      print("downloading, try later")
+            self.gui.mapTiles.addTileDownloadRequest(layerId, z, x, y)
+            #      log.debug("downloading, try later")
             return False
 
-
-class MapLayers(object):
-    def __init__(self, gui):
-        self.gui = gui
-
-        self._wrappedLayers = None
-        # why are wee keeping our own dictionary of wrapped
-        # layers and not just returning a newly wrapped object on demand ?
-        # -> because PySide (1.1.1) segfaults if we don't hold any reference
-        # on the object returned :)
-
-    @property
-    def wrappedLayers(self):
-        # make sure the wrapped layer dict has benn initialized
-        # (we can't do that at init, as at that time the
-        # map layers module is not yet loaded)
-        if self._wrappedLayers is None:
-            self._wrappedLayers = {}
-            #for layerId, layer in six.iteritems(self.gui._mapLayers.getLayerDict()):
-            #    self.wrappedLayers[layerId] = wrappers.MapLayerWrapper(layer)
-        return self._wrappedLayers
-
-    def getLayer(self, layerId):
-        return self.wrappedLayers.get(layerId, None)
-
-    def getLayerName(self, layerId):
-        layer = self.wrappedLayers.get(layerId, None)
-        if layer:
-            return layer.wo.label
-        else:
-            return "label for %s unknown" % layerId
-
-class Search(object):
+class _Search(object):
     _addressSignal = signal.Signal()
 
     changed = signal.Signal()
@@ -445,7 +623,6 @@ class Search(object):
 
     def _threadStatusCB(self, threadName, threadStatus):
         if threadName == self._addressSearchThreadName:
-        #if threadName == constants.THREAD_ADDRESS_SEARCH:
             self._addressSearchStatus = threadStatus
             self._addressSignal()
 
